package ru.dbotthepony.kstarbound.lua

import com.google.gson.JsonArray
import com.google.gson.JsonElement
import com.google.gson.JsonNull
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import com.sun.jna.Native
import com.sun.jna.Pointer
import com.sun.jna.ptr.LongByReference
import org.apache.logging.log4j.LogManager
import org.lwjgl.system.MemoryUtil
import java.io.Closeable
import java.io.PrintWriter
import java.io.StringWriter
import java.lang.RuntimeException
import java.nio.ByteBuffer
import java.nio.ByteOrder

private fun stringToBuffer(str: String): ByteBuffer {
	val bytes = str.toByteArray(charset = Charsets.UTF_8)
	val buf = ByteBuffer.allocateDirect(bytes.size)
	buf.order(ByteOrder.nativeOrder())
	bytes.forEach(buf::put)
	buf.position(0)

	return buf
}

class LuaState : Closeable {
	private val pointer = LuaJNA.INSTANCE.luaL_newstate() ?: throw OutOfMemoryError("Unable to allocate new LuaState")
	private val nativePointer = Pointer.nativeValue(pointer)

	private var destroyed = false
	private val panicHandler = object : LuaJNA.lua_CFunction {
		override fun invoke(luaState: Pointer): Int {
			println("$this panicked!")
			return 0
		}
	}

	init {
		LuaJNA.INSTANCE.lua_atpanic(pointer, panicHandler)

		LuaJNA.INSTANCE.luaopen_base(pointer)
		storeGlobal("_G")
		LuaJNA.INSTANCE.luaopen_package(pointer)
		storeGlobal("package")
		LuaJNA.INSTANCE.luaopen_table(pointer)
		storeGlobal("table")
		LuaJNA.INSTANCE.luaopen_coroutine(pointer)
		storeGlobal("coroutine")
		LuaJNA.INSTANCE.luaopen_string(pointer)
		storeGlobal("string")
		LuaJNA.INSTANCE.luaopen_math(pointer)
		storeGlobal("math")
		LuaJNA.INSTANCE.luaopen_utf8(pointer)
		storeGlobal("utf8")
		LuaJNA.INSTANCE.luaopen_debug(pointer)
		storeGlobal("debug")

		/*
		pushClosure {
			val build = mutableListOf<String?>()

			for (i in 1 .. stackTop) {
				build.add(getString(i))
			}

			LOGGER.info("Lua/print(): {}", build.joinToString("\t"))
		}

		LuaJNA.lua_setglobal(pointer, "print")

		 */
	}

	fun pushClosure(lambda: (state: LuaState) -> Unit) {
		LuaJNI.lua_pushcclosure(nativePointer) lazy@{
			try {
				lambda.invoke(this@LuaState)
			} catch (err: Throwable) {
				val builder = StringWriter()
				val printWriter = PrintWriter(builder)
				err.printStackTrace(printWriter)
				push(builder.toString())
				return@lazy 1
			}

			return@lazy 0
		}
	}

	fun checkStack(minAmount: Int): Boolean {
		return LuaJNA.INSTANCE.lua_checkstack(pointer, minAmount) > 0
	}

	fun absoluteIndex(index: Int): Int {
		return LuaJNA.INSTANCE.lua_absindex(pointer, index)
	}

	fun load(code: String, chunkName: String = "main chunk") {
		val buf = stringToBuffer(code)

		throwLoadError(LuaJNA.INSTANCE.lua_load(pointer, object : LuaJNA.lua_Reader {
			override fun readNextChunk(luaState: Pointer, userdata: Pointer?, size: LongByReference): Pointer? {
				if (buf.remaining() == 0) {
					size.value = 0
					return Pointer.NULL
				}

				size.value = buf.remaining().toLong()
				val p = MemoryUtil.memAddress(buf)
				buf.position(buf.remaining())
				return Pointer(p)
			}
		}, Pointer.NULL, chunkName, "t"))
	}

	fun pcall(numArgs: Int = 0, numResults: Int = 0): Int {
		val status = LuaJNA.INSTANCE.lua_pcallk(pointer, numArgs, numResults, 0, null, null)

		if (status == LUA_ERRRUN) {
			throw LuaRuntimeException(getString())
		}

		return status
	}

	fun getString(index: Int = -1, limit: Long = 4096): String? {
		val len = LongByReference()
		val p = LuaJNA.INSTANCE.lua_tolstring(pointer, absoluteIndex(index), len) ?: return null

		if (len.value == 0L) {
			return ""
		}

		if (len.value >= limit) {
			throw IllegalStateException("Unreasonably long Lua string: ${len.value}")
		}

		val readBytes = ByteArray(len.value.toInt())
		p.read(0L, readBytes, 0, readBytes.size)
		return readBytes.toString(charset = Charsets.UTF_8)
	}

	override fun close() {
		if (destroyed) {
			throw IllegalStateException("Already destroyed")
		}

		LuaJNA.INSTANCE.lua_close(pointer)
		destroyed = true
	}

	val stackTop: Int get() {
		val value = LuaJNA.INSTANCE.lua_gettop(pointer)
		check(value >= 0) { "Invalid stack top $value" }
		return value
	}

	fun storeGlobal(name: String) {
		LuaJNA.INSTANCE.lua_setglobal(pointer, name)
	}

	fun loadGlobal(name: String) {
		LuaJNA.INSTANCE.lua_getglobal(pointer, name)
	}

	fun push() {
		LuaJNA.INSTANCE.lua_pushnil(pointer)
	}

	fun push(value: Int) {
		LuaJNA.INSTANCE.lua_pushinteger(pointer, value.toLong())
	}

	fun push(value: Long) {
		LuaJNA.INSTANCE.lua_pushinteger(pointer, value)
	}

	fun push(value: String) {
		val bytes = value.toByteArray(Charsets.UTF_8)
		val block = Native.malloc(bytes.size.toLong())

		if (block == 0L)
			throw OutOfMemoryError("Unable to allocate ${bytes.size} bytes on heap")

		try {
			Pointer(block).write(0L, bytes, 0, bytes.size)
			LuaJNA.INSTANCE.lua_pushlstring(pointer, block, bytes.size.toLong())
		} finally {
			Native.free(block)
		}
	}

	fun push(value: JsonElement) {
		when (value) {
			JsonNull.INSTANCE -> {
				LuaJNA.INSTANCE.lua_pushnil(pointer)
			}

			is JsonPrimitive -> {
				if (value.isNumber) {
					val num = value.asNumber

					when (num) {
						is Int, is Long -> LuaJNA.INSTANCE.lua_pushinteger(pointer, num.toLong())
						else -> LuaJNA.INSTANCE.lua_pushnumber(pointer, num.toDouble())
					}
				} else if (value.isString) {
					push(value.asString)
				} else if (value.isBoolean) {
					LuaJNA.INSTANCE.lua_pushboolean(pointer, if (value.asBoolean) 1 else 0)
				} else {
					throw IllegalArgumentException(value.toString())
				}
			}

			is JsonArray -> {
				LuaJNA.INSTANCE.lua_createtable(pointer, value.size(), 0)
				val index = stackTop

				for ((i, v) in value.withIndex()) {
					LuaJNA.INSTANCE.lua_pushinteger(pointer, i.toLong() + 1L)
					push(v)

					LuaJNA.INSTANCE.lua_settable(pointer, index)
				}
			}

			is JsonObject -> {
				LuaJNA.INSTANCE.lua_createtable(pointer, 0, value.size())
				val index = stackTop

				for ((k, v) in value.entrySet()) {
					push(k)
					push(v)

					LuaJNA.INSTANCE.lua_settable(pointer, index)
				}
			}

			else -> {
				throw IllegalArgumentException(value.toString())
			}
		}
	}
}
